package com.aliyun.tablestore.grid;

import com.alicloud.openservices.tablestore.*;
import com.alicloud.openservices.tablestore.core.ErrorCode;
import com.alicloud.openservices.tablestore.model.GetRowResponse;
import com.alicloud.openservices.tablestore.model.Row;
import com.alicloud.openservices.tablestore.model.TableOptions;
import com.alicloud.openservices.tablestore.model.UpdateTableRequest;
import com.alicloud.openservices.tablestore.model.search.CreateSearchIndexRequest;
import com.alicloud.openservices.tablestore.model.search.IndexSchema;
import com.alicloud.openservices.tablestore.model.search.SearchResponse;
import com.alicloud.openservices.tablestore.model.search.query.Query;
import com.alicloud.openservices.tablestore.writer.WriterConfig;
import com.aliyun.tablestore.grid.core.TableStoreDataDeleter;
import com.aliyun.tablestore.grid.core.RequestBuilder;
import com.aliyun.tablestore.grid.core.RowParser;
import com.aliyun.tablestore.grid.core.TableStoreDataFetcher;
import com.aliyun.tablestore.grid.core.TableStoreDataWriter;
import com.aliyun.tablestore.grid.model.GridDataSetMeta;
import com.aliyun.tablestore.grid.model.QueryGridDataSetResult;
import com.aliyun.tablestore.grid.model.QueryParams;
import ucar.nc2.ft.point.standard.TableConfig;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

public class TableStoreGrid implements GridStore {

    private final TableStoreGridConfig config;
    private final AsyncClientInterface asyncClient;
    private ExecutorService writerExecutor;
    private TableStoreWriter writer;
    private TableOptions tableOptions;

    public TableStoreGrid(TableStoreGridConfig config) {
        this.config = config;
        this.asyncClient = new AsyncClient(config.getTableStoreEndpoint(),
                config.getAccessId(), config.getAccessKey(),
                config.getTableStoreInstance());
    }

    @Override
    public void createStore() throws Exception {
        this.tableOptions = new TableOptions(-1, 1);
        createStore(tableOptions);
    }
    @Override
    public void createStore(TableOptions tableOptions) throws Exception {
        this.tableOptions = tableOptions;
        // create buffer table
        try {
            tableOptions.setAllowUpdate(true); // 数据表必须允许更新，否则无法导入数据
            this.asyncClient.createTable(RequestBuilder.buildCreateDataTableRequest(config.getDataTableName(), tableOptions), null).get();
        } catch (TableStoreException ex) {
            if (!ex.getErrorCode().equals(ErrorCode.OBJECT_ALREADY_EXIST)) {
                throw ex;
            }
        }

        // create meta table
        try {
            tableOptions.setAllowUpdate(false); // 属性表为了设置过期时间，禁止update操作，只允许put操作
            this.asyncClient.createTable(RequestBuilder.buildCreateMetaTableRequest(config.getMetaTableName(), tableOptions), null).get();
        } catch (TableStoreException ex) {
            if (!ex.getErrorCode().equals(ErrorCode.OBJECT_ALREADY_EXIST)) {
                throw ex;
            }
        }


    }

    /**
     * 更新表设置
     * @param tableOptionsForUpdate
     * @throws Exception
     */
    public void updateStoreOption(TableOptions tableOptionsForUpdate) throws Exception {
        // create meta table
        try {
            UpdateTableRequest metaUpdate = new UpdateTableRequest(config.getMetaTableName());
            metaUpdate.setTableOptionsForUpdate(tableOptionsForUpdate);
            this.asyncClient.updateTable(metaUpdate, null).get();
        } catch (TableStoreException ex) {
            if (!ex.getErrorCode().equals(ErrorCode.OBJECT_ALREADY_EXIST)) {
                throw ex;
            }
        }

        // create buffer table
        try {
            UpdateTableRequest dataUpdate = new UpdateTableRequest(config.getDataTableName());
            dataUpdate.setTableOptionsForUpdate(tableOptionsForUpdate);
            this.asyncClient.updateTable(dataUpdate, null).get();
        } catch (TableStoreException ex) {
            if (!ex.getErrorCode().equals(ErrorCode.OBJECT_ALREADY_EXIST)) {
                throw ex;
            }
        }
    }


    @Override
    public void putDataSetMeta(GridDataSetMeta meta) throws Exception {
        this.asyncClient.putRow(
                RequestBuilder.buildPutMetaRequest(config.getMetaTableName(), meta), null).get();
    }

    public void deleteDataSetMeta(String uniqueKey) throws Exception {
        this.asyncClient.deleteRow(
                RequestBuilder.buildDeleteMetaRequest(config.getMetaTableName(), uniqueKey),
                null
        );
    }

    @Override
    public void updateDataSetMeta(GridDataSetMeta meta) throws Exception {
        this.asyncClient.updateRow(
                RequestBuilder.buildUpdateMetaRequest(config.getMetaTableName(), meta), null).get();
    }

    @Override
    public GridDataSetMeta getDataSetMeta(String gridDataSetId) throws Exception {
        GetRowResponse getRowResponse = this.asyncClient.getRow(
                RequestBuilder.buildGetMetaRequest(config.getMetaTableName(), gridDataSetId), null).get();
        if (getRowResponse.getRow() == null) {
            return null;
        }
        return RowParser.parseMetaFromRow(getRowResponse.getRow());
    }

    @Override
    public void createMetaIndex(String indexName, IndexSchema indexSchema) throws Exception {
        CreateSearchIndexRequest request = new CreateSearchIndexRequest();
        request.setTableName(config.getMetaTableName());
        request.setIndexName(indexName);
        request.setIndexSchema(indexSchema);
        request.setTimeToLive(tableOptions.getTimeToLive(), TimeUnit.SECONDS);
        this.asyncClient.createSearchIndex(request, null).get();
    }

    @Override
    public QueryGridDataSetResult queryDataSets(String indexName, Query query, QueryParams queryParams) throws Exception {
        SearchResponse searchResponse = this.asyncClient.search(
                RequestBuilder.buildSearchRequest(config.getMetaTableName(), indexName, query, queryParams),
                null).get();
        List<GridDataSetMeta> metaList = new ArrayList<GridDataSetMeta>();
        for (Row row : searchResponse.getRows()) {
            metaList.add(RowParser.parseMetaFromRow(row));
        }
        QueryGridDataSetResult result = new QueryGridDataSetResult();
        result.setGridDataSetMetas(metaList);
        result.setAllSuccess(searchResponse.isAllSuccess());
        result.setNextToken(searchResponse.getNextToken());
        result.setTotalCount(searchResponse.getTotalCount());
        return result;
    }

    @Override
    public GridDataWriter getDataWriter(GridDataSetMeta meta) {
        if (writer == null) {
           synchronized (this) {
               if (writer == null) {
                   this.writerExecutor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
                   this.writer = new DefaultTableStoreWriter(this.asyncClient, config.getDataTableName(), new WriterConfig(), null, this.writerExecutor);
               }
           }
        }
        return new TableStoreDataWriter(writer, config.getDataTableName(), meta);
    }

    @Override
    public GridDataFetcher getDataFetcher(GridDataSetMeta meta) {
        return new TableStoreDataFetcher(asyncClient, config.getDataTableName(), meta, config.getDataSizeLimitForFetch());
    }

    public GridDataDeleter getDataDeleter(GridDataSetMeta meta) {
        return new TableStoreDataDeleter(asyncClient, config.getDataTableName(), meta);
    }
    @Override
    public synchronized void close() {
        if (writer != null) {
            this.writer.close();
            this.writerExecutor.shutdown();
        }
        this.asyncClient.shutdown();
    }


}
