package com.htfp.weather.griddata.operation;

import com.aliyun.tablestore.grid.GridDataFetcher;
import com.aliyun.tablestore.grid.consts.AttributionEnum;
import com.aliyun.tablestore.grid.model.GridDataSet;

import com.aliyun.tablestore.grid.model.GridDataSetMeta;
import com.aliyun.tablestore.grid.model.grid.Range;
import com.htfp.weather.griddata.common.TableConfig;
import com.htfp.weather.griddata.common.ValueRange;
import com.htfp.weather.download.gfs.GfsLevelsEnum;
import com.htfp.weather.utils.NdArrayUtils;
import com.htfp.weather.web.exception.AppException;
import com.htfp.weather.web.exception.ErrorCode;
import lombok.extern.slf4j.Slf4j;
import org.springframework.context.annotation.DependsOn;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;
import ucar.ma2.Array;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import java.time.Duration;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

/**
 * @Author : shiyi
 * @Date : 2024/1/22 13:53
 * @Description : 根据维度信息获取数据, 将物理坐标转换为数组索引，与数据库交互
 */
@Slf4j @Component
@DependsOn({"tableStoreConf", "tableConfig"})
public class GfsDataFetcher extends BaseTableOperation {
    @Resource
    TableConfig tableConfig;
    @Resource
    QueryMeta queryMeta;
    GridDataSetMeta lastGridDataSetMeta;
    @PreDestroy
    public void destroy() {
        close();
    }
    public GridDataSetMeta getLastGridDataSetMeta() {
        if (lastGridDataSetMeta == null) {
            updateLastGridDataSetMeta();
            throw new AppException(ErrorCode.DATA_SET_EMPTY);
        }
        return lastGridDataSetMeta;
    }

    /**
     * 每小时更新一次元数据
     */
    @Scheduled(cron = "0 30 * * * ?") @PostConstruct
    public GridDataSetMeta updateLastGridDataSetMeta() {
        try {
            lastGridDataSetMeta = queryMeta.getLastGridDataSetMeta();
        } catch (Exception e) {
            lastGridDataSetMeta = null;
        }
        log.info("更新最新的数据元信息: {}", lastGridDataSetMeta);
        return lastGridDataSetMeta;
    }

    /**
     * 获取指定变量的四维数据(time, lev, lat, lon)
     * @param dataSetId 数据id
     * @param variables 变量名列表
     * @param origin 维度起始索引
     * @param shape 数据形状
     * @return {@link GridDataSet}
     * @throws Exception
     */
    public GridDataSet queryByTableStore(String dataSetId, List<String> variables, int[] origin, int[] shape) throws Exception {
        GridDataFetcher fetcher = tableStoreGrid.getDataFetcher(tableStoreGrid.getDataSetMeta(dataSetId));
        fetcher.setVariablesToGet(variables);
        fetcher.setOriginShape(origin, shape);
        GridDataSet gridDataSet = fetcher.fetch();
        // Grid4D grid4D = fetcher.fetch().getVariable(variable);
        return gridDataSet;
    }

    public GridDataSet getSinglePoint(String dataSetId, List<String> variables, int iTime, int iLev, int iLat, int iLon) throws Exception {
        int[] origin = new int[] {iTime, iLev, iLat, iLon};
        int[] shape = new int[] {1, 1, 1, 1};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, variables,
                origin, shape);
        return gridDataSet;
    }

    /**
     *  获取指定变量的二维全平面数据
     * @param variable 变量名
     * @param iLev 高度索引
     * @return {@link Array}
     * @throws Exception
     */
    public Array getPlane(String dataSetId, String variable, int iTime, int iLev) throws Exception {
        int[] origin = new int[] {iTime, iLev, 0, 0};
        int[] shape = new int[] {1, 1, tableConfig.latSize, tableConfig.lonSize};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, Collections.singletonList(variable),
                origin, shape);
        Array array = gridDataSet.getVariable(variable).toArray();
        System.out.println("Shape: " + array.shapeToString());
        System.out.println("Data: " + array);
        return array;
    }

    /**
     *  获取给定变量的二维平面范围数据
     * @param variable 变量名
     * @param iLev 高度索引
     * @param iLonStart 起始经度索引
     * @param iLatLength 纬度格点数量
     * @param iLonLength 经度格点数量
     * @return {@link Array}
     * @throws Exception
     */
    public Array getPlane(String dataSetId, String variable, int iTime, int iLev, int iLatStart, int iLonStart, int iLatLength, int iLonLength) throws Exception {
        int[] origin = new int[] {iTime, iLev, iLatStart, iLonStart};
        int[] shape = new int[] {1, 1, iLatLength, iLonLength};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, Collections.singletonList(variable),
                origin, shape);
        Array array = gridDataSet.getVariable(variable).toArray();
        System.out.println("Shape: " + array.shapeToString());
        System.out.println("Data: " + array);
        return array;
    }

    /**
     * 获取指定变量在某位置随气压的分布
     * @param variableNameList 变量名列表
     * @param targetTime 目标时刻
     * @param latitude 经度
     * @param longitude 纬度
     * @return {@link GridDataSet}
     * @throws Exception
     */
    public GridDataSet getProfileByPressure(List<String> variableNameList, OffsetDateTime targetTime, double latitude, double longitude) throws Exception {
        lastGridDataSetMeta = getLastGridDataSetMeta();
        GridDataFetcher fetcher = tableStoreGrid.getDataFetcher(lastGridDataSetMeta);
        int iTime = getTargetTimeIndex(targetTime);
        int iLat= getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        int[] origin = new int[]{iTime, 0, iLat, iLon};
        int[] shape = new int[]{1,tableConfig.levSize, 1, 1};
        // TODO 2024/6/17:
        fetcher.setVariablesToGet(variableNameList);
        fetcher.setOriginShape(origin, shape);
        return fetcher.fetch();
    }

    /**获取变量随近地面高度的分布（目前只针对风速风向）*/
    public GridDataSet getProfileByNearSurfaceHeight(List<String> variableNameList, OffsetDateTime targetTime, double latitude, double longitude) throws Exception {
        lastGridDataSetMeta = getLastGridDataSetMeta();
        GridDataFetcher fetcher = tableStoreGrid.getDataFetcher(lastGridDataSetMeta);
        int iTime = getTargetTimeIndex(targetTime);
        int iLat= getLatitudeIndex(latitude);
        int iLon = getLongitudeIndex(longitude);
        int[] origin = new int[]{iTime, 0, iLat, iLon};
        //TODO NOTE 2024/7/4: 目前仅有风速支持近地面多个高度, 且只有10，20，30，40，50，80，100这7个高度，2m高度只针对temp和humidity；nc文件本身高度坐标就有很多种，难以统一到数据库中
        int[] shape = new int[]{1,tableConfig.heightList.length - 1, 1, 1}; // heightList.length - 1 去掉2m高度
        // TODO 2024/6/17:
        fetcher.setVariablesToGet(variableNameList);
        fetcher.setOriginShape(origin, shape);
        return fetcher.fetch();
    }

    /**
     * 获取指定变量在某经纬高处的预报序列
     * @param dataSetId 数据集ID
     * @param variable 变量名
     * @param iLev     高度索引
     * @param iLat     纬度索引
     * @param iLon     经度索引
     * @throws Exception
     */
    public Array getSeries(String dataSetId, String variable, int[] iTimeList, int iLev, int iLat, int iLon) throws Exception {
        int[] origin = new int[] {iTimeList[0], iLev, iLat, iLon};
        int[] shape = new int[] {iTimeList.length, 1, 1, 1};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, Collections.singletonList(variable), origin, shape);
        Array array = gridDataSet.getVariable(variable).toArray();
        System.out.println("Shape: " + array.shapeToString());
        System.out.println("Data: " + array);
        return array;
    }

    /**
     * 获取指定变量在某位置随时间的变化
     * @param dataSetId 数据集ID
     * @param variableList 变量名列表
     * @param iTImes 时间索引数组
     * @param iLev     高度索引
     * @param iLat     纬度索引
     * @param iLon     经度索引
     * @return
     * @throws Exception
     */
    public GridDataSet getSeries(String dataSetId, List<String> variableList, int[] iTImes, int iLev, int iLat, int iLon) throws Exception {
        int[] origin = new int[] {iTImes[0], iLev, iLat, iLon};
        int[] shape = new int[] {iTImes.length, 1, 1, 1};
        return queryByTableStore(dataSetId, variableList, origin, shape);
    }

    /**
     * 查询指定范围的4维数组数据集
     * @param variableNames
     * @param timeRange
     * @param levRange
     * @param latRange
     * @param lonRange
     * @return
     * @throws Exception
     */
    public GridDataSet queryDataset4D(List<String> variableNames, ValueRange<OffsetDateTime> timeRange, ValueRange<Integer> levRange, ValueRange<Double> latRange, ValueRange<Double> lonRange) throws Exception {
        lastGridDataSetMeta = getLastGridDataSetMeta();
        GridDataFetcher fetcher = tableStoreGrid.getDataFetcher(lastGridDataSetMeta);
        Range iTimeRange = new Range(getTargetTimeIndex(timeRange.getStart()),getTargetTimeIndex(timeRange.getEnd()) + 1);
        Range iLevRage = new Range(getLevelIndex(levRange.getStart()), getLevelIndex(levRange.getEnd()) + 1);
        Range iLatRange = new Range(getLatitudeIndex(latRange.getStart()),getLatitudeIndex(latRange.getEnd()) + 1);
        Range iLonRange = new Range(getLongitudeIndex(lonRange.getStart()),getLongitudeIndex(lonRange.getEnd()) + 1);
        int[] origin = new int[]{iTimeRange.getStart(), iLevRage.getStart(), iLatRange.getStart(), iLonRange.getStart()};
        int[] shape = new int[]{iTimeRange.getSize(), iLevRage.getSize(), iLatRange.getSize(), iLonRange.getSize()};
        // TODO 2024/6/17:
        fetcher.setVariablesToGet(variableNames);
        fetcher.setOriginShape(origin, shape);
        GridDataSet gridDataSet = fetcher.fetch();
        return gridDataSet;
    }


    private int getTargetTimeIndex(OffsetDateTime targetTime) {
        lastGridDataSetMeta = getLastGridDataSetMeta();
        long milli = (long) lastGridDataSetMeta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        if (targetTime.isBefore(refTime)) {
            throw new AppException(ErrorCode.QUERY_TIME_ERROR);
        }
        List<String> forecastHours = lastGridDataSetMeta.getForecastHours();
        String targetHour = String.valueOf(getForecastHourFromTargetTime(targetTime, refTime));
        if (!forecastHours.contains(targetHour)) {
            throw new AppException(ErrorCode.QUERY_TIME_ERROR);
        }
        int iTime = forecastHours.indexOf(targetHour);
        return iTime;
    }

    public int getLongitudeIndex(double lon) {
        double[] lonList = tableConfig.getLonList();
        if (lonList.length == 0) {
            throw new AppException(ErrorCode.LONGITUDE_INDEX_ERROR, "经度列表为空， 无法查找索引");
        }
        double minLon = lonList[0];
        double maxLon = lonList[lonList.length - 1];
        if (minLon > lon || lon > maxLon) {
            throw new AppException(ErrorCode.LONGITUDE_INDEX_ERROR, "经度不在有效范围内: " + minLon + " ~ " + maxLon);
        }
        int nearestIndex = NdArrayUtils.findNearestIndex(lonList, lon);
        if (nearestIndex == -1) {
            throw new AppException(ErrorCode.LONGITUDE_INDEX_ERROR, "找不到" + lon + "的经度索引");
        }
        return nearestIndex;
    }

    public int getLatitudeIndex(double lat) {
        double[] latList = tableConfig.getLatList();
        if (latList.length == 0) {
            throw new AppException(ErrorCode.LATITUDE_INDEX_ERROR, "纬度列表为空， 无法查找索引");
        }
        double minLat = latList[0];
        double maxLat = latList[latList.length - 1];

        if (minLat > lat || lat > maxLat) {
            throw new AppException(ErrorCode.LATITUDE_INDEX_ERROR, "纬度不在有效范围内: " + minLat + " ~ " + maxLat);
        }
        int nearestIndex = NdArrayUtils.findNearestIndex(latList, lat);
        if (nearestIndex == -1) {
            throw new AppException(ErrorCode.LATITUDE_INDEX_ERROR, "找不到" + lat + "的纬度索引");
        }
        return nearestIndex;
    }

    private int getPressureIndex(int pressureLevel) {
        List<Integer> levList = Arrays.stream(tableConfig.getPressureList()).boxed().collect(Collectors.toList());
        if (CollectionUtils.isEmpty(levList)) {
            throw new AppException(ErrorCode.LEVEL_INDEX_ERROR, "气压列表为空， 无法查找索引");
        }
        int index = levList.indexOf(pressureLevel);
        if (index == -1) {
            throw new AppException(ErrorCode.LATITUDE_INDEX_ERROR, "找不到" + pressureLevel + "的气压高度索引");
        }
        return index;
    }

    public int getLevelIndex(int level) {
        if (GfsLevelsEnum.getByCode(level) == GfsLevelsEnum.SURFACE) {
            return 0; // NOTE 2024/5/20: 0 表示地面，风速对应10m高度，其他变量对应2m高度
        } else {
            return getPressureIndex(level);
        }
    }

    private int getForecastHourFromTargetTime(OffsetDateTime targetTime, OffsetDateTime refTime) {
        return (int) Duration.between(refTime, targetTime).toHours();
    }

}
