package com.htfp.weather.utils;

import java.time.*;
import java.time.format.DateTimeFormatter;
import java.util.TimeZone;

/**
 * @Author : shiyi
 * @Date : 2024/1/23 18:18
 * @Description : 日期时间工具类
 */
public class DateTimeUtils {
    public static final String DEFAULT_PATTERN = "yyyy-MM-dd'T'HH:mm:ssxxx";
    public static final DateTimeFormatter DEFAULT_FORMATTER = DateTimeFormatter.ofPattern(DEFAULT_PATTERN);
    public static final ZoneOffset DEFAULT_ZONE_OFFSET = ZoneOffset.ofHours(TimeZone.getDefault().getRawOffset()/3600000);
    public static final ZoneId DEFAULT_ZONE_ID = ZoneId.systemDefault();
    /**
     * 将某时区的时间字符串更改为其他时区的时间字符串,
     *
     * @param dateTimeStr 输入的时间字符串，需包含时区信息，如+08:00
     * @param offsetTo    输出时间的时区，如8代表东八区
     * @return 更改时区之后的时间字符串，输出格式与输入格式一致
     */
    public static String convertZonedDateTime(String dateTimeStr, int offsetTo, String formatStr) {
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(formatStr);
        // 转换为UTC时区的ZonedDateTime对象，时间点保持一致
        OffsetDateTime offsetDateTime = OffsetDateTime.parse(dateTimeStr, formatter)
                .withOffsetSameInstant(ZoneOffset.ofHours(offsetTo));
        return offsetDateTime.format(formatter);
    }

    /**
     * 以指定的格式，将某时区的时间字符串更改为其他时区的时间字符串
     *
     * @param dateTimeStr 输入的时间字符串，不需要包含时区信息，如果含有时区信息将会被忽略
     * @param zoneFrom    输入时间的时区, dateTimeStr中的时区信息将被忽略
     * @param zoneTo      输出时间的时区，如8代表东八区
     * @param formatStr   输入和输出的时间字符串格式，可以不包含时区信息
     * @return 指定时区表示的时间字符串
     */
    public static String convertZonedDateTime(String dateTimeStr, int zoneFrom, int zoneTo, String formatStr) {
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(formatStr);
        ZoneId zoneIdFrom = ZoneId.ofOffset("", ZoneOffset.ofHours(zoneFrom));
        ZoneId zoneIdTo = ZoneId.ofOffset("", ZoneOffset.ofHours(zoneTo));

        // LocalDateTime解析将会忽略dateTimeStr中的时区信息
        // OffsetDateTime.parse(dateTimeStr)
        ZonedDateTime zonedDateTime = LocalDateTime.parse(dateTimeStr, formatter)
                .atZone(zoneIdFrom)
                .withZoneSameInstant(zoneIdTo);
        return zonedDateTime.format(formatter);
    }


    public static OffsetDateTime getUTCDateTime(OffsetDateTime offsetDateTime) {
        return offsetDateTime.withOffsetSameInstant(ZoneOffset.UTC);
    }

    // public static LocalDateTime getUTCDateTime(String timeString, String formatStr) {
    //     DateTimeFormatter formatter = new DateTimeFormatterBuilder()
    //             .appendPattern(formatStr)
    //             .toFormatter();
    //
    //     // 转换为UTC时区的ZonedDateTime对象，时间点保持一致
    //     ZonedDateTime utcZonedDateTime = LocalDateTime.parse(timeString, formatter)
    //             .atZone(ZoneId.systemDefault())
    //             .withZoneSameInstant(ZoneId.of("UTC"));
    //     return utcZonedDateTime.toLocalDateTime();
    // }

    public static ZonedDateTime getLocalZoneDateTime(OffsetDateTime utcDateTime) {
        ZonedDateTime zonedDateTime = utcDateTime.toZonedDateTime().withZoneSameInstant(ZoneId.systemDefault());
        return zonedDateTime;
    }
    /**
     * 将UTC时间字符串转换为将北京时时间字符串
     *
     * @param utcTimeStr UTC时间字符串
     * @param formatStr  输入和输出的字符串格式
     * @return 北京时（东八区）时间字符串
     */
    public static String getBJTDateTimeStringFromUTC(String utcTimeStr, String formatStr) {
        return convertZonedDateTime(utcTimeStr, 0, 8, formatStr);
    }

    /**
     * 将北京时时间字符串转换为UTC时间字符串
     *
     * @param bjtTimeStr  北京时（东八区）时间字符串
     * @param formatStr  输入和输出的字符串格式
     * @return UTC时间字符串
     */
    public static String getUTCDateTimeStringFromBJT(String bjtTimeStr, String formatStr) {
        return convertZonedDateTime(bjtTimeStr, 8, 0, formatStr);
    }

    /** 获取指定日期的时间戳（毫秒）
     * @param dateTimeStr 时区时间字符串
     * @param formatStr 指定日期字符串的格式
     * @return 毫秒时间戳
     */
    public static long getTimestampMilli(String dateTimeStr, String formatStr) {
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(formatStr);
        return OffsetDateTime.parse(dateTimeStr, formatter).toInstant().toEpochMilli();
    }
    public static long getTimestampMilli(String dateTimeStr) {
        return OffsetDateTime.parse(dateTimeStr, DEFAULT_FORMATTER).toInstant().toEpochMilli();
    }
    /**
     * 将时间戳转换为时间字符串
     * @param timestampMilli
     * @param zoneOffset
     * @return yyyy-MM-dd'T'HH:mmxxx 格式字符串
     */
    public static String getTimeStringFromMilli(long timestampMilli, int zoneOffset) {
        Instant instant = Instant.ofEpochMilli(timestampMilli);
        // 使用 ZonedDateTime 类将 Instant 对象转换为本地时区时间
        ZonedDateTime zonedDateTime = instant.atZone(ZoneId.ofOffset("", ZoneOffset.ofHours(zoneOffset)));
        return zonedDateTime.format(DEFAULT_FORMATTER);
    }
    public static String getTimeStringFromMilli(long timestampMilli, String zoneId) {
        Instant instant = Instant.ofEpochMilli(timestampMilli);
        // 使用 ZonedDateTime 类将 Instant 对象转换为本地时区时间
        ZonedDateTime zonedDateTime = instant.atZone(ZoneId.of(zoneId));
        return zonedDateTime.format(DEFAULT_FORMATTER);
    }
    /**
     * 将时间戳转换为本地时区的字符串
     * @param timestampMilli
     * @return yyyy-MM-dd'T'HH:mmxxx 格式字符串
     */
    public static String getTimeStringFromMilli(long timestampMilli) {
        Instant instant = Instant.ofEpochMilli(timestampMilli);
        // 使用 ZonedDateTime 类将 Instant 对象转换为本地时区时间
        ZonedDateTime zonedDateTime = instant.atZone(ZoneId.systemDefault());
        return zonedDateTime.format(DEFAULT_FORMATTER);
    }

    /** 将OffsetDateTime对象转换为指定时区*/
    public static OffsetDateTime offsetDateTimeToTargetZone(OffsetDateTime offsetDateTime, int ZoneTo) {
        return offsetDateTime.withOffsetSameInstant(ZoneOffset.ofHours(ZoneTo));
    }
    /** 将OffsetDateTime对象转换为系统时区*/
    public static OffsetDateTime offsetDateTimeToSystemZone(OffsetDateTime offsetDateTime) {
        return offsetDateTime.withOffsetSameInstant(DEFAULT_ZONE_OFFSET);
    }
}
