package com.htfp.weather.utils;

import com.htfp.weather.utils.ssl.HostInterceptor;
import com.htfp.weather.utils.ssl.SelectiveHostnameVerifier;
import com.htfp.weather.utils.ssl.SelectiveTrustManager;
import lombok.extern.slf4j.Slf4j;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.Request.Builder;
import okhttp3.Response;
import okhttp3.ResponseBody;
import org.apache.commons.lang3.StringUtils;
import org.springframework.util.CollectionUtils;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.nio.file.Files;
import java.nio.file.StandardCopyOption;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;

import javax.net.ssl.*;
import javax.net.ssl.SSLContext;
import java.security.SecureRandom;

/**
 * @Author : shiyi
 * @Date : 2024/1/23 16:19
 * @Description : HttpClient工具类
 */
@Slf4j
public class HttpClientUtils {
    private static final List<String> TRUSTED_DOMAINS = Arrays.asList("nomads.ncep.noaa.gov");
    private static final OkHttpClient OKHTTP_CLIENT = new OkHttpClient.Builder()
            .connectTimeout(30, TimeUnit.SECONDS)
            .readTimeout(60, TimeUnit.SECONDS)
            .writeTimeout(60, TimeUnit.SECONDS)
            .build();
    private static final OkHttpClient DOWNLOAD_OKHTTP_CLIENT = buildOKHttpClient(TRUSTED_DOMAINS).build();

    public static OkHttpClient.Builder buildOKHttpClient(List<String> trustedDomains) {
        try {
            // 1. 创建自定义 TrustManager
            SelectiveTrustManager trustManager = new SelectiveTrustManager(trustedDomains);
            // 2. 初始化 SSLContext
            SSLContext sslContext = SSLContext.getInstance("TLS");
            sslContext.init(null, new TrustManager[]{trustManager}, new SecureRandom());
            final SSLSocketFactory sslSocketFactory = sslContext.getSocketFactory();
            OkHttpClient.Builder builder = new OkHttpClient.Builder();
            builder.sslSocketFactory(sslSocketFactory, trustManager);
            builder.addInterceptor(new HostInterceptor());
            builder.hostnameVerifier(new SelectiveHostnameVerifier(trustedDomains));
            return builder;
        } catch (NoSuchAlgorithmException | KeyManagementException e) {
            e.printStackTrace();
            return new OkHttpClient.Builder();
        } catch (Exception e) {
            e.printStackTrace();
            return new OkHttpClient.Builder();
        }
    }

    // public static <T> T sendGet(String url, Map<String, String> params, Class<T> responseClass) {
    //     MultiValueMap<String, String> queryParams = new LinkedMultiValueMap<>();
    //     if (params != null) {
    //         params.forEach(queryParams::add);
    //     }
    //     Mono<T> mono = WebClient.create(url).method(HttpMethod.GET)
    //             .uri(uriBuilder -> uriBuilder
    //                     .queryParams(queryParams)
    //                     .build()).retrieve().bodyToMono(responseClass);
    //     T responseBody;
    //     try {
    //         responseBody = mono.block();
    //     } catch (WebClientResponseException e) {
    //         log.error("Get 请求错误, {}", e.getMessage());
    //         return null;
    //     }
    //     return responseBody;
    // }
    public static <T> T sendGet(String url, Map<String, String> params, Class<T> responseClass) throws Exception {
        String jsonStr = sendGet(url, params);
        return JSONUtils.json2obj(jsonStr, responseClass);
    }

    public static String sendGet(String url, Map<String, String> params) throws IOException {
        StringBuilder stringBuilder = new StringBuilder();
        if (!CollectionUtils.isEmpty(params)) {
            for (String res : params.keySet()) {
                if (res == null || params.get(res) == null) {
                    continue;
                }
                if (StringUtils.isNotBlank(stringBuilder)) {
                    stringBuilder.append("&");
                } else {
                    stringBuilder.append("?");
                }
                try {
                    stringBuilder.append(String.format("%s=%s", res, URLEncoder.encode(params.get(res), "UTF-8")));
                } catch (UnsupportedEncodingException e) {
                    throw new RuntimeException(e);
                }
            }
        }
        Request request = new Builder()
                .url(url + stringBuilder)
                .build();

        try (Response response = OKHTTP_CLIENT.newCall(request).execute()) {
            if (!response.isSuccessful()) {
                throw new IOException("okhttp Unexpected code " + response);
            }
            ResponseBody body = response.body();
            if (body != null) {
                return body.string();
            } else {
                return null;
            }
        }
    }


    public static void downloadFileByUrl(String sourceUrl, String localPath) throws IOException {
        File file = new File(localPath);
        Request request = new Request.Builder()
                .url(sourceUrl)
                .build();

        try (Response response = DOWNLOAD_OKHTTP_CLIENT.newCall(request).execute()) {
            if (response.code() == 200) {
                assert response.body() != null;
                InputStream stream = response.body().byteStream();
                long dataBytes = Files.copy(stream, file.toPath(), StandardCopyOption.REPLACE_EXISTING);
                log.debug("文件{}下载完成，文件大小: {} bytes", localPath, dataBytes);
            } else {
                throw new RuntimeException("文件下载失败，Response：" + response);
            }
        }

    }

}
