package com.htfp.weather.utils;

import ucar.ma2.Array;
import ucar.ma2.ArrayFloat;

import java.util.Arrays;

/**
 * @Author : shiyi
 * @Date : 2024/1/24 11:32
 * @Description : 地理信息与气象数据处理工具类
 */
public class MeteoUtils {

    /**
     * 风速单位转换
     *
     * @param mps 风速 (m/s)
     * @return 风速 (km/h)
     */
    public static double mPerSecond2kmPerHour(double mps) {
        return mps * 3.6;
    }

    /**
     * @param kph 风速 (km/h)
     * @return 风速 (m/s)
     */
    public static double kmPerHour2mPerSecond(double kph) {
        return kph / 3.6;
    }

    /**
     * 风力等级转换
     * @param kph–风速 (km/h)
     * @return 蒲福风力等级
     */
    public static int kmPerHour2WindScale(double kph) {
        return mPerSecond2WindScale(kmPerHour2mPerSecond(kph));
    }
    /**
     * 风力等级转换
     * @param mps–风速 (m/s)
     * @return 蒲福风力等级
     */
    public static int mPerSecond2WindScale(double mps) {
        if (mps < 0) {
            throw new IllegalArgumentException("风力等级计算错误风速不能为负值");
        }
        if (mps <= 0.2) {
            return 0;
        } else if (mps <= 1.5) {
            return 1;
        } else if (mps <= 3.3) {
            return 2;
        } else if (mps <= 5.4) {
            return 3;
        } else if (mps <= 7.9) {
            return 4;
        } else if (mps <= 10.7) {
            return 5;
        } else if (mps <= 13.8) {
            return 6;
        } else if (mps <= 17.1) {
            return 7;
        } else if (mps <= 20.7) {
            return 8;
        } else if (mps <= 24.4) {
            return 9;
        } else if (mps <= 28.4) {
            return 10;
        } else if (mps <= 32.6) {
            return 11;
        } else if (mps <= 36.9){
            return 12;
        } else {
            return 13;
        }
    }


    /**
     * 根据标准状况根据气压计算海拔高度（不考虑温度），仅供参考
     * @param currentPressure 当前大气压（hPa）
     * @return 高度
     */
    public static double pressure2Elevation(double currentPressure) {
        double referencePressure = 101.325; // 参考大气压（单位：kPa）
        return 44330 * (1 - Math.pow(currentPressure/10. / referencePressure, 1 / 5.255));
    }
    /**
     * @param array 降水率 (kg/m^2/s)
     * @return mm/h
     */
    public static Array precipRate2mmPerHour(Array array) {
        for (int i = 0; i < array.getSize(); i++) {
            array.setFloat(i, array.getFloat(i)*3600);
        }
        return array;
    }
    /**
     * 开尔文转摄氏度
     */
    public static Array kelvin2Celsius(Array array) {
        for (int i = 0; i < array.getSize(); i++) {
            array.setFloat(i, array.getFloat(i) - 273.15f);
        }
        return array;
    }
    public static double calculateWindSpeed(double u, double v) {
        return Math.sqrt(u * u + v * v);
    }
    //
    // @Deprecated
    // public static Array calculateWindSpeed(Array u, Array v) {
    //     // note 2024/5/13: io耗时太多，对于一般大小的矩阵不实用
    //     uvValid(u, v);
    //     try (NDManager manager = NDManager.newBaseManager()) {
    //         int[] shapeArray = u.getShape();
    //         Shape shape = new Shape(Arrays.stream(shapeArray).mapToLong(i -> (long) i).toArray());
    //         NDArray uwnd = manager.create((float[]) u.getStorage(), shape);
    //         NDArray vwnd = manager.create((float[]) v.getStorage(), shape);
    //         NDArray windSpeed = (uwnd.square().add(vwnd.square())).sqrt();
    //         ByteBuffer byteBuffer = windSpeed.toByteBuffer();
    //         return Array.factory(u.getDataType(), shapeArray, byteBuffer);
    //     } catch (Exception e) {
    //         throw e;
    //     }
    // }

    private static void uvValid(Array u, Array v) {
        if (u.getSize() != v.getSize()) {
            throw new IllegalArgumentException("u 和 v 数据量不一致");
        }
        if (u.getSize() == 0 || v.getSize() == 0) {
            throw new IllegalArgumentException("u 和 v 数据不能为空");
        }
        if (!Arrays.equals(u.getShape(), v.getShape())) {
            throw new IllegalArgumentException("u 和 v 数据形状不一致");
        }
        if (!u.getDataType().equals(v.getDataType())) {
            throw new IllegalArgumentException("u 和 v 数据类型不一致");
        }
    }

    public static Array calculateWindSpeed(Array u, Array v) {
        uvValid(u, v);
        Array windSpeed = new ArrayFloat(u.getShape());
        for (int i = 0; i < windSpeed.getSize(); i++) {
            windSpeed.setFloat(i, (float) calculateWindSpeed(u.getFloat(i), v.getFloat(i)));
        }
        return windSpeed;

    }

    public static Array calculateWindDirection(Array u, Array v) {
        uvValid(u, v);
        Array windDirection = new ArrayFloat(u.getShape());
        for (int i = 0; i < windDirection.getSize(); i++) {
            windDirection.setFloat(i, (float) calculateWindDirection(u.getFloat(i), v.getFloat(i)));
        }
        return windDirection;

    }
    public static double calculateWindDirection(double u, double v) {
        // DecimalFormat df = new DecimalFormat("#.00");

        double windDirDeg;

        if (nearToZero(u) && v > 0.) {
            windDirDeg = 180;
        } else if (nearToZero(u) && v < 0.) {
            windDirDeg = 0;
        } else if (u > 0. && nearToZero(v)) {
            windDirDeg = 270;
        } else if (u < 0. && nearToZero(v)) {
            windDirDeg = 90;
        } else if (nearToZero(u) && nearToZero(v)) {
            windDirDeg = 0;
        } else {
            windDirDeg = calculateWindDirection0(u, v);
        }
        return (float) windDirDeg;
    }

    public static double calculateWindDirection0(double u, double v) {
        // DecimalFormat df = new DecimalFormat("#.00");
        double windDirRad = Math.atan2(v, u);
        return (270 - windDirRad * 180 / Math.PI) % 360;
    }

    private static boolean nearToZero(double value) {
        return Math.abs(value) < 1e-5;
    }
}
