package com.htfp.weather.web.service;

import com.aliyun.tablestore.grid.consts.AttributionEnum;
import com.aliyun.tablestore.grid.model.GridDataSet;
import com.aliyun.tablestore.grid.model.GridDataSetMeta;
import com.aliyun.tablestore.grid.model.grid.Grid4D;
import com.htfp.weather.download.DownLoadFileInfo;
import com.htfp.weather.download.gfs.GfsDownloader;
import com.htfp.weather.download.gfs.GfsLevelsEnum;
import com.htfp.weather.download.gfs.GfsVariableHeightEnum;
import com.htfp.weather.download.gfs.GfsVariableIsobaricEnum;
import com.htfp.weather.griddata.common.TableConfig;
import com.htfp.weather.griddata.common.ValueRange;
import com.htfp.weather.griddata.operation.GfsDataFetcher;
import com.htfp.weather.griddata.operation.GfsDataImport;
import com.htfp.weather.info.Constant;
import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.utils.MeteoUtils;
import com.htfp.weather.web.exception.AppException;
import com.htfp.weather.web.exception.ErrorCode;
import com.htfp.weather.web.param.response.*;
import com.htfp.weather.web.service.surfaceapi.ISurfaceDataService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import ucar.ma2.Array;
import ucar.ma2.Index4D;

import javax.annotation.Resource;
import javax.annotation.security.DenyAll;
import java.lang.reflect.Method;
import java.time.Instant;
import java.time.OffsetDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @Author : shiyi
 * @Date : 2024/2/2 15:47
 * @Description : 获取tablestore格点数据
 */
@Service("tablestore-gfs")
@Slf4j
public class GfsDataServiceImpl implements IUpperDataService, ISurfaceDataService {
    @Resource
    GfsDataFetcher gfsDataFetcher;
    @Resource
    TableConfig tableConfig;
    @Resource
    GfsDownloader gfsDownloader;
    @Resource
    GfsDataImport gfsDataImport;

    @Override
    public String getDataSource() {
        return dataSource;
    }

    String dataSource = "GFS";
    // 地面站数据访问接口
    @Override
    public NowWeatherStatus getNowSurfaceWeatherStatus(double lat, double lon) throws Exception {
        return getNowWeather(GfsLevelsEnum.SURFACE.getCode(), lat, lon);
    }

    @Override
    public TimeSeriesDataset getForecastSeries(double lat, double lon) throws Exception {
        return getForecastSeries(GfsLevelsEnum.SURFACE.getCode(), lat, lon);
    }

    @Override @DenyAll
    public List<SurfaceWeatherWarning> getSurfaceWarning(double lat, double lon) throws Exception {
        return null;
    }


    // 下载和导入数据
    public DownloadResult downloadNowAllFile() {
        gfsDownloader.iniTimeSetting();
        List<DownLoadFileInfo> downLoadFileInfoList = gfsDownloader.getFilesInfo();
        List<DownLoadFileInfo> finishedList = gfsDownloader.downloadAll(downLoadFileInfoList);
        List<DownLoadFileInfo> successList = finishedList.stream().filter(DownLoadFileInfo::isDownloadSuccess).collect(Collectors.toList());
        List<DownLoadFileInfo> failedList = finishedList.stream().filter(fileInfo -> !fileInfo.isDownloadSuccess()).collect(Collectors.toList());
        DownloadResult downloadResult = new DownloadResult();
        downloadResult.setSuccessList(successList);
        downloadResult.setFailedList(failedList);
        return downloadResult;
    }

    public DownLoadFileInfo downloadSingleFile(OffsetDateTime offsetDateTime) {
        String targetUtcStr = DateTimeUtils.getUTCDateTime(offsetDateTime).format(DateTimeFormatter.ofPattern(Constant.UTC_TIME_STRING));
        try {
            gfsDownloader.iniTimeSetting();
            List<DownLoadFileInfo> downLoadFileInfoList = gfsDownloader.getFilesInfo(); // TODO 2024/6/12: 可以优化，不用生成24小时所有的文件信息
            for (DownLoadFileInfo downLoadFileInfo : downLoadFileInfoList) {
                if (downLoadFileInfo.getForecastUTCTimeStr().equals(targetUtcStr)) {
                    DownLoadFileInfo downloadResult = gfsDownloader.download(downLoadFileInfo);
                    if (downloadResult.isDownloadSuccess()) {
                        return downloadResult;
                    } else {
                        throw new AppException(ErrorCode.DOWNLOAD_ERROR, downloadResult.getErrorMsg());
                    }
                }
            }
            throw new AppException(ErrorCode.QUERY_TIME_ERROR, ": 请选择当前时间之后24小时范围内的时刻");
        } catch (Exception e) {
            throw new AppException(ErrorCode.DOWNLOAD_START_ERROR, e.getMessage());
        }
    }

    public ImportResult importData(OffsetDateTime time) throws Exception {
        List<GfsDataImport.ImportFileInfo> finishedList = gfsDataImport.importData(time);
        List<GfsDataImport.ImportFileInfo> successList = finishedList.stream().filter(result -> result.isSuccess()).collect(Collectors.toList());
        List<GfsDataImport.ImportFileInfo> failedList = finishedList.stream().filter(result -> !result.isSuccess()).collect(Collectors.toList());
        ImportResult importResult = new ImportResult();
        importResult.setFailedList(failedList);
        importResult.setSuccessList(successList);
        return importResult;

    }


    // 从tablestore获取数据
    /**
     * 获取当前时刻的天气状况
     *
     * @param latitude  纬度
     * @param longitude 经度
     * @param level     高度：9999代表地面，其他代表指定气压高度
     * @return @{@link NowWeatherStatus}
     */
    @Override
    public NowWeatherStatus getNowWeather(int level, double latitude, double longitude) {
        OffsetDateTime nowTime = OffsetDateTime.now();
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Collections.singletonList(nowTime));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(Collections.singletonList(latitude));
        ValueRange<Double> lonRange = new ValueRange<>(Collections.singletonList(longitude));
        List<String> variableNames = getVariableNamesInDatabase(level);
        try {
            GridDataSet gridDataSet = gfsDataFetcher.queryDataset4D(
                    variableNames,timeRange, levRange, latRange, lonRange
            );
            final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
            // 每个变量都是0维数组
            NowWeatherStatus nowWeatherStatus = buildNowWeatherInTargetPoint(gridDataSet, level, latitude, longitude, levelFlag);
            nowWeatherStatus.setLatitude(latitude);
            nowWeatherStatus.setLongitude(longitude);
            return nowWeatherStatus;
        } catch (AppException e) {
            throw e;
        }catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }
    }

    /**
     * 获得指定点序列上的当前气象信息
     * @param latitude
     * @param longitude
     * @param level
     * @return
     */
    @Override
    public List<NowWeatherStatus> getNowWeatherByMultiPoint(int level, List<Double> latitude, List<Double> longitude) {
        if (latitude.size() != longitude.size()) {
            throw new AppException(ErrorCode.INDEX_SIZE_ERROR, ": 经纬度数组大小不一致");
        }
        int pointSize = latitude.size();
        List<String> variableNames = getVariableNamesInDatabase(level);
        OffsetDateTime nowTime = OffsetDateTime.now();
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Collections.singletonList(nowTime));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(latitude);
        ValueRange<Double> lonRange = new ValueRange<>(longitude);
        try {
            GridDataSet gridDataSet = gfsDataFetcher.queryDataset4D(
                    variableNames,timeRange, levRange, latRange, lonRange
            );
            List<NowWeatherStatus> nowWeatherStatusList = new ArrayList<>();

            for (int i = 0; i < pointSize; i++) {
                NowWeatherStatus nowWeatherStatus;
                // NOTE 2024/6/11: build中传入的索引和数据库的查询索引意义不同
                final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
                nowWeatherStatus = buildNowWeatherInTargetPoint(gridDataSet, level, latitude.get(i), longitude.get(i), levelFlag);
                nowWeatherStatusList.add(nowWeatherStatus);
            }
            return nowWeatherStatusList;
        } catch (AppException e) {
            throw e;
        }catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }
    }

    /**
     * 获取未来24小时的气象时间序列数据
     *
     * @param latitude  纬度
     * @param longitude 经度
     * @param level     高度：9999代表地面，其他代表指定气压高度
     * @return @{@link TimeSeriesDataset}
     */
    @Override
    public TimeSeriesDataset getForecastSeries(int level, double latitude, double longitude) {
        List<String> variableNames = getVariableNamesInDatabase(level);
        OffsetDateTime nowTime = OffsetDateTime.now();
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Arrays.asList(nowTime, nowTime.plusHours(24)));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(Collections.singletonList(latitude));
        ValueRange<Double> lonRange = new ValueRange<>(Collections.singletonList(longitude));

        try {
            GridDataSet gridDataSet = gfsDataFetcher.queryDataset4D(
                    variableNames, timeRange, levRange, latRange, lonRange
            );
            final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
            return buildTimeSeriesDatasetInTargetPoint(gridDataSet, level, latitude, longitude, levelFlag);
        } catch (AppException e) {
            throw e;
        }catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }
    }

    /**
     * 获得指定点序列上的未来24小时气象预报
     * @param latitudeList
     * @param longitudeList
     * @param level
     * @return
     */
    @Override
    public List<TimeSeriesDataset> getForecastSeriesByMultiPoint(int level, List<Double> latitudeList, List<Double> longitudeList) {
        if (latitudeList.size() != longitudeList.size()) {
            throw new AppException(ErrorCode.INDEX_SIZE_ERROR, ": 经纬度数组大小不一致");
        }
        int pointSize = latitudeList.size();
        List<String> variableNames = getVariableNamesInDatabase(level);
        OffsetDateTime nowTime = OffsetDateTime.now();
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Arrays.asList(nowTime, nowTime.plusHours(24)));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(latitudeList);
        ValueRange<Double> lonRange = new ValueRange<>(longitudeList);
        try {
            GridDataSet gridDataSet = gfsDataFetcher.queryDataset4D(
                    variableNames, timeRange, levRange, latRange, lonRange
            );
            List<TimeSeriesDataset> timeSeriesDatasetList = new ArrayList<>();
            for (int i = 0; i < pointSize; i++) {
                final GfsLevelsEnum levelFlag = GfsLevelsEnum.getByCode(level);
                TimeSeriesDataset timeSeriesDataset = buildTimeSeriesDatasetInTargetPoint(gridDataSet, level, latitudeList.get(i), longitudeList.get(i), levelFlag);
                timeSeriesDatasetList.add(timeSeriesDataset);
            }
            return timeSeriesDatasetList;
        } catch (AppException e) {
            throw e;
        }catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }
    }

    /**
     * 获取变量随气压变化的廓线
     *
     * @param targetTime   目标时间 UTC
     * @param variableName 数据库中的变量名
     * @param latitude     纬度
     * @param longitude    经度
     * @return {@link ProfileResponse}
     */
    public ProfileResponse getProfileByVariableAndPressure(String variableName, OffsetDateTime targetTime, double latitude, double longitude) {
        String targetVariable = GfsVariableIsobaricEnum.getGfsVariableName(variableName);
        if (targetVariable == null) {
            throw new AppException(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }
        try {
            GridDataSet gridDataSet = gfsDataFetcher.getProfileByPressure(Collections.singletonList(targetVariable), targetTime, latitude, longitude);
            float[] values = (float[]) gridDataSet.getVariables().get(targetVariable).toArray().copyTo1DJavaArray();
            ProfileResponse profileResponse = new ProfileResponse();
            profileResponse.setPressureLevels(tableConfig.getPressureList());
            profileResponse.setHeightLevels(tableConfig.getPressureHeightList());
            profileResponse.setValues(values);
            profileResponse.setLatitude(latitude);
            profileResponse.setLongitude(longitude);
            return profileResponse;
        } catch (AppException e) {
            throw e;
        }catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }
    }

    /** 获取全部变量随气压变化的廓线 (推荐) */
    @Override
    public ProfileDataset getProfileByPressure(OffsetDateTime targetTime, double latitude, double longitude) {
        try {
            GfsLevelsEnum levelFlag = GfsLevelsEnum.UPPER;
            List<String> variableNames = getVariableNamesInDatabase(levelFlag.getCode());
            GridDataSet gridDataSet = gfsDataFetcher.getProfileByPressure(variableNames, targetTime, latitude, longitude);
            ProfileDataset profile = buildProfileDataset(gridDataSet, levelFlag);
            profile.setLatitude(latitude);
            profile.setLongitude(longitude);
            return profile;
        } catch (AppException e) {
            throw e;
        } catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }
    }

    /** 获取指定变量随高度变化的廓线， 目前只支持风速风向*/
    public ProfileResponse getProfileByVariableAndNearSurfaceHeight(String variableName, OffsetDateTime targetTime, double latitude, double longitude) {
        String targetVariable = GfsVariableHeightEnum.getGfsVariableName(variableName);
        if (targetVariable == null) {
            throw new AppException(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }
        try {
            int[] heightLevels = tableConfig.getHeightList();
            GridDataSet gridDataSet = gfsDataFetcher.getProfileByNearSurfaceHeight(Collections.singletonList(targetVariable), targetTime, latitude, longitude);
            float[] values = (float[]) gridDataSet.getVariables().get(targetVariable).toArray().copyTo1DJavaArray();
            ProfileResponse profileResponse = new ProfileResponse();
            profileResponse.setValues(values);
            profileResponse.setHeightLevels(Arrays.copyOfRange(heightLevels, 1, heightLevels.length));
            profileResponse.setLatitude(latitude);
            profileResponse.setLongitude(longitude);
            return profileResponse;
        } catch (AppException e) {
            throw e;
        }catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }
    }

    /** 获取全部变量随高度变化的廓线， 目前只支持风速风向*/
    public ProfileDataset getProfileByNearSurfaceHeight(OffsetDateTime targetTime, double latitude, double longitude) {
        try {
            GfsLevelsEnum levelFlag = GfsLevelsEnum.SURFACE;
            List<String> variableNames = new ArrayList<>();
            variableNames.add(GfsVariableHeightEnum.getGfsVariableName("windSpeed"));
            variableNames.add(GfsVariableHeightEnum.getGfsVariableName("wind360"));
            GridDataSet gridDataSet = gfsDataFetcher.getProfileByNearSurfaceHeight(variableNames, targetTime, latitude, longitude);
            ProfileDataset profile = buildProfileDataset(gridDataSet, levelFlag);
            profile.setHeightLevel(Arrays.copyOfRange(tableConfig.getHeightList(), 1, tableConfig.getHeightList().length));
            profile.setLatitude(latitude);
            profile.setLongitude(longitude);
            return profile;
        } catch (AppException e) {
            throw e;
        } catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }
    }

    /**
     * 获取指定变量、时间、高度的二维数组数据
     *
     * @param targetTime   目标时间 UTC
     * @param variableName api中的变量名
     * @param level        高度
     * @param minLat       最小纬度
     * @param maxLat       最大纬度
     * @param minLon       最小经度
     * @param maxLon       最大经度
     * @return @{@link PlaneResponse}
     */
    @Override
    public PlaneResponse getPlane(OffsetDateTime targetTime, String variableName, int level, double minLat, double maxLat, double minLon, double maxLon) {
        String targetVariable;
        if (GfsLevelsEnum.SURFACE.equals(GfsLevelsEnum.getByCode(level))) {
            targetVariable = GfsVariableHeightEnum.getGfsVariableName(variableName);
        } else {
            targetVariable = GfsVariableIsobaricEnum.getGfsVariableName(variableName);
        }

        if (targetVariable == null) {
            throw new AppException(ErrorCode.VALIDATE_ERROR, variableName + "变量在数据库中不存在");
        }

        List<String> variableNames = Collections.singletonList(targetVariable);
        ValueRange<OffsetDateTime> timeRange = new ValueRange<>(Collections.singletonList(targetTime));
        ValueRange<Integer> levRange = new ValueRange<>(Collections.singletonList(level));
        ValueRange<Double> latRange = new ValueRange<>(minLat, maxLat);
        ValueRange<Double> lonRange = new ValueRange<>(minLon, maxLon);

        try {
            Array array = gfsDataFetcher.queryDataset4D(
                    variableNames, timeRange, levRange, latRange, lonRange
            ).getVariable(targetVariable).toArray();
            return buildPlane(array, minLat, maxLat, minLon, maxLon);
        } catch (AppException e) {
            throw e;
        }catch (Exception e) {
            throw new AppException(ErrorCode.TABLESTORE_QUERY_ERROR, e.getMessage());
        }

    }


    private List<String> getVariableNamesInDatabase(int level) {
        List<String> variableNames = new ArrayList<>();
        if (GfsLevelsEnum.SURFACE.equals(GfsLevelsEnum.getByCode(level))) {
            for (String e : tableConfig.getVariableList()) {
                String gfsVariableName = GfsVariableHeightEnum.getGfsVariableName(e);
                if (gfsVariableName != null) {
                    variableNames.add(gfsVariableName);
                }
            }
        } else {
            for (String e : tableConfig.getVariableList()) {
                String gfsVariableName = GfsVariableIsobaricEnum.getGfsVariableName(e);
                if (gfsVariableName != null) {
                    variableNames.add(gfsVariableName);
                }
            }
        }
        return variableNames;
    }

    private PlaneResponse buildPlane(Array array, double minLat, double maxLat, double minLon, double maxLon) {
        int iMinLat = gfsDataFetcher.getLatitudeIndex(minLat);
        int iMaxLat = gfsDataFetcher.getLatitudeIndex(maxLat);
        int iMinLon = gfsDataFetcher.getLongitudeIndex(minLon);
        int iMaxLon = gfsDataFetcher.getLongitudeIndex(maxLon);
        int latLength = iMaxLat - iMinLat + 1;
        int lonLength = iMaxLon - iMinLon + 1;
        PlaneResponse planeResponse = new PlaneResponse();
        float[][] values = (float[][]) array.reduce().copyToNDJavaArray();
        planeResponse.setLatList(Arrays.copyOfRange(tableConfig.getLatList(), iMinLat, iMinLat + latLength));
        planeResponse.setLonList(Arrays.copyOfRange(tableConfig.getLonList(), iMinLon, iMinLon + lonLength));
        planeResponse.setValues(values);
        return planeResponse;
    }


    /**
     * 构建指定位置的预报数据
     * @param gridDataSet
     * @param level
     * @param latitude
     * @param longitude
     * @param levelFlag-高度层次标记，地面或高空
     * @return
     */

    private TimeSeriesDataset buildTimeSeriesDatasetInTargetPoint(GridDataSet gridDataSet, int level, double latitude, double longitude,GfsLevelsEnum levelFlag) {
        TimeSeriesDataset result = new TimeSeriesDataset();
        result.setLatitude(latitude);
        result.setLongitude(longitude);
        // 获取全局索引
        int levIndexGlobal = gfsDataFetcher.getLevelIndex(level);
        int latIndexGlobal = gfsDataFetcher.getLatitudeIndex(latitude);
        int lonIndexGlobal = gfsDataFetcher.getLongitudeIndex(longitude);

        // 利用反射构建数据集
        Class<? extends TimeSeriesDataset> aClass = result.getClass();
        for (Map.Entry<String, Grid4D> entry : gridDataSet.getVariables().entrySet()) {
            String variableNameInApi = getVariableNameInApi(levelFlag, entry);
            Grid4D grid4D = entry.getValue();
            int[] origin = grid4D.getOrigin(); // grid4D的起始坐标全局索引
            int[] shape = grid4D.getShape();
            int tSize = shape[0];
            try {
                String setMethodName = "set" + variableNameInApi.substring(0, 1).toUpperCase() + variableNameInApi.substring(1);
                Method method =  aClass.getDeclaredMethod(setMethodName, aClass.getDeclaredField(variableNameInApi).getType());
                // TODO 2024/6/20: 这里的toArray操作在数据量较大的时候会比较吃内存和cpu
                float[] storage = (float[]) grid4D.toArray().section(
                        // levIndexGlobal-origin[1] 获取目标点在子区域中的索引
                        new int[]{0, levIndexGlobal-origin[1], latIndexGlobal-origin[2], lonIndexGlobal-origin[3]}, new int[]{tSize, 1, 1, 1}
                ).copyTo1DJavaArray();
                method.invoke(result, storage);
            } catch (ReflectiveOperationException e) {
                throw new AppException(ErrorCode.NO_SUCH_VARIABLE, e);
            } catch (Exception e) {
                e.printStackTrace();
                throw new AppException(ErrorCode.RESPONSE_DATA_BUILD_ERROR, e);
            }

            // 设置时间
            if (result.getTime() == null) {
                List<String> timeList = getResultTimeStringList(gridDataSet, tSize, origin[0]);
                result.setTime(timeList);
            }
            // 计算地面风力等级
            if (levelFlag == GfsLevelsEnum.SURFACE) {
                float[] windSpeed = result.getWindSpeed();
                if (windSpeed != null) {
                    int[] windScale = new int[windSpeed.length];
                    for (int i = 0; i < windScale.length; i++) {
                        windScale[i] = MeteoUtils.mPerSecond2WindScale(windSpeed[i]);
                    }
                    result.setWindScale(windScale);
                }
            }
        }
        return result;
    }


    private List<String> getResultTimeStringList(GridDataSet gridDataSet, int timeSize, int iTimeStart) {
        GridDataSetMeta meta = gridDataSet.getMeta();
        List<String> forecastHours = meta.getForecastHours();
        long milli = (long) meta.getAttributes().get(AttributionEnum.REFERENCE_TIME.getName());
        OffsetDateTime refTime = OffsetDateTime.ofInstant(Instant.ofEpochMilli(milli), ZoneOffset.ofHours(0));
        List<String> timeList = new ArrayList<>();
        for (int i = 0; i < timeSize; i++) {
            int hour = Integer.parseInt(forecastHours.get(i + iTimeStart));
            String timeStr = DateTimeUtils.getLocalZoneDateTime(refTime.plusHours(hour))
                    .format(Constant.API_TIME_FORMATTER);
            timeList.add(timeStr);
        }
        return timeList;
    }

    /**
     * 构建指定位置的当前天气数据
     * @param gridDataSet
     * @param level
     * @param latitude
     * @param longitude
     * @param levelFlag-高度层次标记，地面或高空
     * @return
     */
    private NowWeatherStatus buildNowWeatherInTargetPoint(GridDataSet gridDataSet, int level, double latitude, double longitude,  GfsLevelsEnum levelFlag) {
        String timeStr = OffsetDateTime.now().withMinute(0).withNano(0).format(Constant.API_TIME_FORMATTER);
        NowWeatherStatus result = new NowWeatherStatus();
        result.setTime(timeStr);
        result.setLatitude(latitude);
        result.setLongitude(longitude);
        // 获取全局索引
        int levIndexGlobal = gfsDataFetcher.getLevelIndex(level);
        int latIndexGlobal = gfsDataFetcher.getLatitudeIndex(latitude);
        int lonIndexGlobal = gfsDataFetcher.getLongitudeIndex(longitude);

        // 利用反射构建数据集
        Class<? extends NowWeatherStatus> aClass = result.getClass();
        for (Map.Entry<String, Grid4D> entry : gridDataSet.getVariables().entrySet()) {
            String variableNameInApi = getVariableNameInApi(levelFlag, entry);
            Grid4D grid4D = entry.getValue(); Index4D index = new Index4D(grid4D.getShape());
            int[] origin = grid4D.getOrigin(); // grid4D的起始坐标全局索引
            try {
                String setMethodName = "set" + variableNameInApi.substring(0, 1).toUpperCase() + variableNameInApi.substring(1);
                Method method =  aClass.getDeclaredMethod(setMethodName, aClass.getDeclaredField(variableNameInApi).getType());
                // TODO 2024/6/19: 避免toArray操作，减少内存复制
                method.invoke(result, grid4D.toArray().getFloat(
                        index.set(0, levIndexGlobal-origin[1], latIndexGlobal-origin[2], lonIndexGlobal-origin[3])
                ));
            } catch (ReflectiveOperationException e) {
                throw new AppException(ErrorCode.NO_SUCH_VARIABLE, e);
            } catch (Exception e) {
                e.printStackTrace();
                throw new AppException(ErrorCode.RESPONSE_DATA_BUILD_ERROR);
            }
        }
        if (levelFlag == GfsLevelsEnum.SURFACE) {
            if (result.getWindSpeed() != null) {
                result.setWindScale(MeteoUtils.mPerSecond2WindScale(result.getWindSpeed()));
            }
        }
        return result;
    }

    private ProfileDataset buildProfileDataset(GridDataSet gridDataSet, GfsLevelsEnum levelFlag) {
        ProfileDataset result = new ProfileDataset();
        // 利用反射构建数据集
        Class<? extends ProfileDataset> aClass = result.getClass();
        for (Map.Entry<String, Grid4D> entry : gridDataSet.getVariables().entrySet()) {
            try {
                String variableNameInApi = getVariableNameInApi(levelFlag, entry);
                Grid4D grid4D = entry.getValue();
                int[] shape = grid4D.getShape();
                int zSize = shape[1];
                String setMethodName = "set" + variableNameInApi.substring(0, 1).toUpperCase() + variableNameInApi.substring(1);
                Method method =  aClass.getDeclaredMethod(setMethodName, aClass.getDeclaredField(variableNameInApi).getType());
                // TODO 2024/6/20: 这里的toArray操作在数据量较大的时候会比较吃内存和cpu
                float[] values = (float[]) grid4D.toArray().section(
                        new int[]{0, 0, 0, 0}, new int[]{1, zSize, 1, 1}
                ).copyTo1DJavaArray();
                method.invoke(result, values);
            } catch (ReflectiveOperationException e) {
                throw new AppException(ErrorCode.NO_SUCH_VARIABLE, e);
            } catch (Exception e) {
                e.printStackTrace();
                throw new AppException(ErrorCode.RESPONSE_DATA_BUILD_ERROR);
            }
        }
        if (levelFlag == GfsLevelsEnum.UPPER) {
            result.setPressureLevels(tableConfig.getPressureList());
            result.setHeightLevel(tableConfig.getPressureHeightList());
        } else {
            result.setHeightLevel(tableConfig.getHeightList());
        }
        return result;
    }

    private String getVariableNameInApi(GfsLevelsEnum levelFlag, Map.Entry<String, Grid4D> variable) {
        String variableNameInFile = variable.getKey();
        String variableNameInApi;
        if (GfsLevelsEnum.SURFACE.equals(levelFlag)) {
            variableNameInApi = GfsVariableHeightEnum.getVariableNameInApi(variableNameInFile);
        } else {
            variableNameInApi = GfsVariableIsobaricEnum.getVariableNameInApi(variableNameInFile);
        }
        if (variableNameInApi == null) {
            throw new AppException(ErrorCode.NO_SUCH_VARIABLE);
        }
        return variableNameInApi;
    }

}
